#
# Copyright (C) 2002 Jp Calderone.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
""" IRC info bot """

# twisted imports
from twisted.persisted import dirdbm, styles
from twisted.protocols import irc, basic
from twisted.internet import protocol, reactor, app, defer
from twisted.python import rebuild, reflect, log, plugin, components
from twisted.scripts import twistd
from twisted import copyright, python

# system imports
import gc, re, os, sys, string, time, md5, random, copy, traceback
import operator, getpass, types

try:
    import cStringIO as StringIO
except:
    import StringIO

try:
    import cPickle as pickle
except:
    import pickle

try:
    from Crypto.PublicKey import RSA
except ImportError:
    RSA = None

# Our friends
import util, info, shorter, secure, logview, interfaces, adapters, error
from interfaces import IAuthenticator
from adapters import ServicesAuthenticator, PasswordAuthenticator
 
randsource = lambda x: ''.join([chr(random.choice(range(256))) for r in range(x)])

# Break lines longer than this into multiple messages
MAX_MSG_LEN = 300

def formatTime(seconds):
    seconds = int(seconds)
    if seconds < 60:
        n = seconds
        u = 'second'
    elif seconds < 60 * 60:
        n = (seconds + 30) / 60
        u = 'minute'
    elif seconds < 60 * 60 * 24:
        n = (seconds + 60 * 30) / (60 * 60)
        u = 'hour'
    else:
        n = (seconds + 60 * 60 * 12) / (60 * 60 * 24)
        u = 'day'
    return 'about %d %s%s' % (n, u, n != 1 and 's' or '')


def deferredGetPlugIns(plugInType, add = [], debugInspection=0, showProgress=0):
    """Asynchronous version of getPlugIns
    
    @rtype: C{Deferred}
    """
    from twisted.internet import threads
    return threads.deferToThread(
        plugin.getPluginFileList, debugInspection, showProgress
    ).addCallback(
        lambda result: threads.deferToThread(
            plugin.loadPlugins,
            plugInType,
            result + add,
            debugInspection,
            showProgress
        )
    )

class User:
    def __init__(self, nick):
        self.nick = nick
        self.channels = util.CaseList()
        self.lastseen = time.time()
    

    def __repr__(self):
        return 'User(%r, %r, %r)' % (self.nick, self.channels, self.lastseen)


class Conversation:
    def __init__(self, user, channel, name, started = None):
        self.recordedBy = user
        self.channel = channel
        self.name = name
        
        if started is None:
            self.started = time.time()
        else:
            self.started = started
        self.ended = None

    def __eq__(self, other):
        if not isinstance(other, Conversation):
            return 0
        return self.channel == other.channel and self.name == other.name


class SecureCommand:
    def __init__(self, method, *args, **kwargs):
        self.method, self.args, self.kwargs = method, args, kwargs
        self.key = ''.join([chr(random.choice(range(256))) for x in range(8)])
        self.created = time.time()


    def getChallenge(self):
        return self.key


    def verifyResponse(self, value):
        if time.time() - self.created > 60:
            return 0
        return self.key == secure.challengeResponse(value)


    def invoke(self):
        self.method(*self.args, **self.kwargs)


class InstantDisconnectProtocol(protocol.Protocol):
    def connectionMade(self):
        self.transport.loseConnection()


class InfoBot(irc.IRCClient, components.Componentized):
    versionName = 'Twisted %s; InfoBot 0.6.3' % (copyright.version,)
    password = 'pynfo'

    PREFIX = '%'

    SRC_URL = "http://www.freshmeat.net/projects/pynfo"

    QUIT_MSGS = [
        'Bye!'
    ]

    debug = 0
    lineRate = 1.0

    LOG_LINES = 500

    hasQuit = 0

    # Things we're discarding
    discards = None
    
    # Recent events which have occurred
    recents = None

    overrideReset = None
    overrideHandler = None

    # Rolling ping time average
    pingAve = 0.0

    # If a filter has yet raised an exception (to prevent excessive log spam)
    filterRaised = 0

    """An IRC info bot."""
    def __init__(self, network, plugins):
        components.Componentized.__init__(self)

        self.network = network
        self.nickname = self.network.nickname
        self.plugins = plugins

        self.logoffFlag = 0
        self.throttle = 30

        self.starttime = time.time()
        self.queries = 0
        self.rebuilds = 0

        self.secureList = []
        self._accum = []
        
        self.discards = {'join': 0, 'part': 0, 'quit': 0}
        self.recents = {'join': [], 'part': [], 'quit': []}

        self.outBytes = 0
        self.inBytes = 0


    def __repr__(self):
        return 'InfoBot(%r, %r)' % (self.nickname, self.network.address)


    def lineReceived(self, line):
        self.inBytes += len(line) + 2
        if self.debug:
            log.msg('lineReceived(%r)' % (line,))
        irc.IRCClient.lineReceived(self, line)
    
    
    def sendLine(self, line):
        self.outBytes += len(line) + 2
        if self.debug:
            log.msg('sendLine(%r)' % (line,))
        irc.IRCClient.sendLine(self, line)


    def connectionMade(self):
        irc.IRCClient.connectionMade(self)
        self.repeatingPing(300)
        log.msg("%s connected to %s" % (self.nickname, self.network.address))


    def handleCommand(self, command, prefix, params):
        if command == self.overrideReset:
            self.overrideHandler = None
        
        if self.overrideHandler:
            self.overrideHandler(command, prefix, params)
        else:
            irc.IRCClient.handleCommand(self, command, prefix, params)


    def signedOn(self):
        map(self.join, self.network.channels)


    def isPrivileged(self, name):
        return self.getPrivilegeLevel(name) != -1


    def getPrivilegeLevel(self, name):
        return self.network.findOwner(name)


    def filterWhoList(self, list):
        return [x[1][5] for x in list if len(x[1]) > 5]


    def getUserList(self, channel):
        return self.who(channel).addCallback(self.filterWhoList)


    def handleModeChange(self, nick, channel, set, mode, args):
        pass


    def irc_RPL_NOSUCHSERVER(self, prefix, params):
        self._accum[0][1].errback(ValueError('no such server'))
        del self._accum[0]


    def irc_RPL_WHOREPLY(self, prefix, params):
        self._accum[0][0].append((prefix, params))


    def irc_RPL_ENDOFWHO(self, prefix, params):
        self._accum[0][1].callback(self._accum[0][0])
        del self._accum[0]


    def authenticated(self, result, pmeth, bot, user, reply, args):
        if result:
            try:
                pmeth(bot, user, reply, args.strip())
            except error.UsageError, e:
                bot.reply(reply, str(e))
            except:
                s = StringIO.StringIO()
                traceback.print_exc(file=s)
                bot.reply(user, "An unexpected error occurred:\n" + s.getvalue())
        else:
            self.reply(reply, 'You are not an identified user.')


    def generateChallenge(self, result, smeth, bot, user, reply, args):
        if result:
            sec = SecureCommand(smeth, bot, user, reply, args)
            challenge = self.factory.encrypt(sec.getChallenge())
            self.reply(user, "Command queued.  Challenge: " + challenge)
            self.secureList.append(sec)
        else:
            self.reply(reply, 'You are not an identified user.')


    def accumulateWhoIsResponse(self, command, prefix, params):
        self._accum[0][0].append((command, prefix, params))

    def irc_RPL_WHOISUSER(self, prefix, params):
        self.overrideReset = 'RPL_ENDOFWHOIS'
        self.overrideHandler = self.accumulateWhoIsResponse
        self.accumulateWhoIsResponse('RPL_WHOISUSER', prefix, params)
        

    def irc_RPL_ENDOFWHOIS(self, prefix, params):
        self._accum[0][1].callback(self._accum[0][0])
        del self._accum[0]


    """def irc_320(self, prefix, params):
        me, nick, res = params
        if self._auth[0][0] == nick:
            self._auth[0][1].callback(1)
            self._auth[0] = 'called'
        else:
            for i in self._auth:
                i[1].callback(0)
            self._auth = []
            log.msg('Authorization error!  Authorization list reset.')
    irc_307 = irc_320"""


    def _bridge(self, channel, user, message):
        for (net, bot) in self.factory.allBots.items():
            if net != self.network.alias and net in self.network.bridged:
                if message.startswith('<action> '):
                    bot.receiveBridgedAction(self.network.alias, user, channel, message)
                else:
                    bot.receiveBridgedMessage(self.network.alias, user, channel, message)


    def bridgeMessage(self, user, channel, msg):
        self._bridge(channel, user, '<%s@%s> %s' % (user, self.network.alias, msg))


    def bridgeAction(self, user, channel, msg):
        self._bridge(channel, user, '* %s@%s %s' % (user, self.network.alias, msg))
    
    
    def checkThrottle(self, channel):
        if self.discards.setdefault(channel, 0):
            return 0
        self.recents.setdefault(channel, []).append(time.time())
        if len(self.recents[channel]) > 5:
            del self.recents[channel][0]
            if self.recents[channel][0] + 1 > self.recents[channel][-1]:
                self.discards[channel] = 1
                reactor.callLater(15, self.discards.__setitem__, channel, 0)
                self.remoteReply(channel, "[%s throttling enabled]" % channel.capitalize())
                return 0
        return 1


    def bridgeJoin(self, user, host, channel):
        if not self.checkThrottle('join'):
            return
        self._bridge(channel, user, '%s (%s) has joined on %s' % (user, host, self.network.alias))


    def bridgePart(self, user, channel):
        if not self.checkThrottle('part'):
            return
        self._bridge(channel, user, '%s@%s has parted' % (user, self.network.alias))

    
    def bridgeKick(self, channel, kicker, kickee, reason='No reason'):
        self._bridge(
            channel, kicker,
            '%s kicked by %s: %s' % (kickee, kicker, reason)
        )

    def bridgeQuit(self, user, msg):
        if not self.checkThrottle('quit'):
            return
        a = self.network.alias
        for (net, bot) in self.factory.allBots.items():
            if net != a and a in bot.network.bridged:
                for c in self.network.users[user].channels:
                    if c in bot.network.channels:
                        bot.receiveBridgedMessage(
                            a, user, c,
                            '%s@%s has quit (%s)' % (
                                user, a, msg
                            )
                        )


    def bridgeNick(self, oldnick, newnick):
        for c in self.network.users[newnick].channels:
            self._bridge(c, newnick, '%s@%s is now known as %s.' % (oldnick, self.network.alias, newnick))


    def bridgeTopic(self, changedBy, channel, newTopic):
        self._bridge(channel, changedBy, 'Topic change on %s by %s: %s' % (self.network.alias, changedBy, newTopic))
    
    
    def bridgeMode(self, nick, channel, set, mode, args):
        self._bridge(
            channel, nick,
            "Mode change on %s by %s: %s%s %s" % (
                self.network.alias, nick, set and '+' or '-', mode, args
            )
        )


    def receiveBridgedAction(self, network, user, channel, msg):
        if channel in self.network.channels:
            irc.IRCClient.me(self, channel, msg[len('<action> '):])


    def receiveBridgedMessage(self, network, user, channel, msg):
        if channel in self.network.channels:
            irc.IRCClient.msg(self, channel, msg)

    def updateUser(self, nick, channel):
        try:
            u = self.network.users[nick.lower()]
        except KeyError:
            u = User(nick)
            u.channels.append(channel)
        else:
            if channel not in u.channels:
                u.channels.append(channel)
        u.lastseen = time.time()
        self.network.users[nick.lower()] = u

    def topicUpdated(self, user, channel, newTopic):
        user = user.split('!')[0]
        if user.lower() != self.nickname.lower():
            self.bridgeTopic(user, channel, newTopic)

    def joined(self, channel):
        self.factory.log(self.network.alias, channel, self.nickname, 'I have joined.')

    def noticed(self, user, channel, message):
        pass

    def userKicked(self, kickee, channel, kicker, message):
        self.bridgeKick(channel, kicker, kickee, message)
        try:
            u = self.network.users[kickee]
        except KeyError:
            pass
        else:
            if channel in u.channels:
                u.channels.remove(channel)
                self.network.users[kickee] = u

    def kickedFrom(self, channel, kicker, message):
        if channel in self.network.channels:
            self.network.channels.remove(channel)
            self.factory.setNetwork(self.network)
            self.join(channel)

    def privmsg(self, user, channel, msg):
        user = string.split(user, '!', 1)[0]
        channel = channel.lower()

        for filter, function in self.plugins['filter_'].iteritems():
            try:
                op = function(self, user, channel, msg)
            except:
                if not self.filterRaised:
                    log.msg("%s raised an exception:" % filter)
                    log.deferr()
                    self.filterRaised = 1
            else:
                if op == 'swallow':
                    return

        self.updateUser(user, channel)
        self.bridgeMessage(user, channel, msg)

        lmsg = msg.lower()
        lnick = self.nickname.lower()
        nickl = len(lnick)

        if channel == lnick:
            self.factory.log(self.network.alias, user, 'pm', msg)
        else:
            self.factory.log(self.network.alias, channel, user, msg)

        if channel == lnick:
            # Turn private queries into a format we can understand
            if not msg.startswith(self.PREFIX):
                msg = self.PREFIX + msg
            elif lmsg.startswith(lnick):
                msg = self.PREFIX + msg[nickl:].lstrip()
            elif lmsg.startswith(lnick) and len(lmsg) > nickl and lmsg[nickl] in string.punctuation:
                msg = self.PREFIX + msg[nickl + 1:].lstrip()
        else:
            # Turn 'ourname:' prefixes into self.PREFIX prefixes
            if lmsg.startswith(lnick) and len(lmsg) > nickl and lmsg[nickl] in string.punctuation:
                msg = self.PREFIX + msg[len(self.nickname) + 1:].lstrip()
        
        # Answer to the right place
        reply = (channel == lnick) and user or channel

        if msg.startswith(self.PREFIX):
            msg = msg[1:]
            parts = msg.lstrip().split(' ', 1)
            if len(parts) == 1:
                parts = parts + ['']
            cmd, args = parts

            if cmd in self.network.disabled:
                self.reply(reply, 'That command has been disabled.')
                return

            ucmd = cmd.upper()
            meth = getattr(self.__class__, 'info_' + ucmd, self.plugins['info_'].get(ucmd, None))
            pmeth = getattr(self.__class__, 'privinfo_' + ucmd, self.plugins['privinfo_'].get(ucmd, None))
            smeth = getattr(self.__class__, 'secure_' + ucmd, self.plugins['secure_'].get(ucmd, None))
            
            try:
                if meth:
                    try:
                        meth(self, user, reply, args.strip())
                    except error.UsageError, e:
                        self.reply(reply, str(e))
                    except:
                        s = StringIO.StringIO()
                        traceback.print_exc(file=s)
                        self.reply(user, "An unexpected error occurred:\n" + s.getvalue())
                    self.queries = self.queries + 1
                elif pmeth:
                    if self.isPrivileged(user):
                        self.getComponent(IAuthenticator).authenticate(user).addCallback(
                            self.authenticated, pmeth, self, user, reply, args
                        ).addErrback(self.unhandledFailure, reply)
                        self.queries = self.queries + 1
                        log.msg("Handling pmeth %s from %s" % (ucmd, user))
                    else:
                        self.reply(reply, "You aren't allowed to do that.")
                elif smeth:
                    if self.isPrivileged(user):
                        self.getComponent(IAuthenticator).authenticate(user).addCallback(
                            self.generateChallenge, smeth, self, user, reply, args
                        ).addErrback(self.unhandledFailure, reply)
                        self.queries = self.queries + 1
                        log.msg("Handling smeth %s from %s" % (ucmd, user))
                    else:
                        self.reply(reply, "You aren't allowed to do that.")
            except:
                f = StringIO.StringIO()
                traceback.print_exc(file=f)
                self.reply(reply, "An error occurred:\n" + f.getvalue())


    def unhandledFailure(self, failure, channel):
        f = StringIO.StringIO()
        failure.printTraceback(file=f)
        self.reply(channel, "An error occurred:\n" + f.getvalue())


    def action(self, user, channel, msg):
        user = string.split(user, '!', 1)[0]
        channel = channel.lower()

        for filter, function in self.plugins['filter_'].iteritems():
            try:
                op = function(self, user, channel, msg)
            except:
                if not self.filterRaised:
                    log.msg("%s raised an exception:" % filter)
                    log.deferr()
                    self.filterRaised = 1
            else:
                if op == 'swallow':
                    return
        self.updateUser(user, channel)
        self.bridgeAction(user, channel, msg)

    #
    # convenience commands
    #
    def who(self, channel):
        self.sendLine('WHO %s' % (channel,))
        self._accum.append(([], defer.Deferred()))
        return self._accum[-1][1]


    def whois(self, name):
        self.sendLine('WHOIS %s' % (name,))
        self._accum.append(([], defer.Deferred()))
        return self._accum[-1][1]


    def reply(self, channel, message, length = MAX_MSG_LEN):
        self.remoteReply(channel, message, length)
        if message.startswith('<action> '):
            self.me(channel, message[len('<action> '):])
        else:
            self.msg(channel, message, length)


    def remoteReply(self, channel, message, length = MAX_MSG_LEN):
        if channel.startswith('#') or channel.startswith('&'):
            self._bridge(channel, self.nickname, message)
            self.factory.log('I', channel, self.nickname, message)


    def pong(self, user, secs):
        self.pingAve = ((self.pingAve * 5) + secs) / 6.0

    def repeatingPing(self, delay):
        reactor.callLater(delay, self.repeatingPing, delay)
        self.ping(self.nickname)


    #
    # irc callbacks
    #
    def modeChanged(self, user, channel, set, modes, args):
        nick = user.split('!', 1)[0]
        self.bridgeMode(nick, channel, set, modes, args)
        self.handleModeChange(nick, channel, set, modes, args)


    def irc_RPL_NAMREPLY(self, prefix, params):
        channel = params[2]
        names = params[3].split()
        for i in names:
            self.updateUser(i, channel)


    def irc_NICK(self, prefix, params):
        """When an IRC user changes their nickname
        """
        old_nick = string.split(prefix,'!')[0]
        new_nick = params[0]
        if old_nick.lower() == self.nickname.lower():
            self.nickname = new_nick
            self.network.nickname = new_nick
            self.factory.setNetwork(self.network)
            for c in self.network.channels:
                self.factory.log(self.network.alias, c, new_nick, "<I am now known as %s>" % (new_nick,))
        elif self.network.users.has_key(old_nick):
            u = copy.copy(self.network.users[old_nick])
            u.nick = new_nick
            self.network.users[old_nick].channels = util.CaseList()
            self.network.users[new_nick] = u
            for c in u.channels:
                self.factory.log(
                    self.network.alias, c, self.nickname,
                    "<%s is now known as %s>" % (old_nick, new_nick)
                )
            self.bridgeNick(old_nick, new_nick)
        else:
            self.network.users[new_nick] = User(new_nick)
        

    def irc_JOIN(self, prefix, params):
        irc.IRCClient.irc_JOIN(self, prefix, params)
        nick, host = prefix.split('!', 1)
        channel = params[0]
        if nick.lower() != self.nickname.lower():
            self.bridgeJoin(nick, host, channel)
            self.updateUser(nick, channel)
        elif channel not in self.network.channels:
            self.network.channels.append(channel)
            self.factory.setNetwork(self.network)


    def irc_PART(self, prefix, params):
        nick = prefix.split('!')[0]
        channel = params[0]
        if nick.lower() != self.nickname.lower():
            self.bridgePart(nick, channel)
            try:
                u = self.network.users[nick]
            except KeyError:
                pass
            else:
                if channel in u.channels:
                    u.channels.remove(channel)
                    self.network.users[nick] = u
        else:
            if channel in self.network.channels:
                self.network.channels.remove(channel)
                self.factory.setNetwork(self.network)
                self.join(channel)
    
    
    def irc_QUIT(self, prefix, params):
        nick = prefix.split('!')[0]
        msg = params[0]
        try:
            u = self.network.users[nick]
        except KeyError:
            pass
        else:
            self.bridgeQuit(nick, msg)
            u.channels = util.CaseList()
            self.network.users[nick] = u

    #
    # InfoBot command methods
    #
    def info_QUERY(self, user, channel, args):
        """Usage: query <user>@<network> <message> - Send a message to a user on a remote network"""
        if channel == user:
            args = args.split(None, 1)
            if len(args) == 1:
                raise error.UsageError
            else:
                target, message = args
                args = target.split('@', 1)
                if len(args) == 1:
                    raise error.UsageError
                else:
                    target, network = args
                    try:
                        bot = self.factory.allBots[network]
                    except KeyError:
                        self.msg(channel, "No such network")
                    else:
                        if bot is self:
                            self.msg(channel, "You could do that yourself.")
                        else:
                            msg = '<%s@%s> %s' % (channel, self.network.alias, message)
                            bot.msg(target, msg)
                            self.msg(channel, "Sent: " + msg)


    def reportWhois(self, lines, user, channel):
        s = ''
        for (command, prefix, params) in lines:
            if command == 'RPL_WHOISUSER':
                s = s + 'Hostmask: ' + params[1] + '!' + params[2] + '@' + params[3] + '\n'
            elif command == 'RPL_WHOISCHANNELS':
                s = s + 'Channels: ' + params[2] + '\n'
            elif command == '320':
                s = s +  user + ' ' + params[2]
        self.reply(channel, s)


    def info_WHOIS(self, user, channel, args):
        """Usage: whois <user>@<network> - Get user info for someone on a bridged network"""
        args = args.split('@', 1)
        if len(args) == 1:
            raise error.UsageError
        else:
            target, network = args
            try:
                bot = self.factory.allBots[network]
            except KeyError:
                self.reply(channel, "No such network")
            else:
                if bot is self:
                    self.reply(channel, "You could do that yourself.")
                else:
                    bot.whois(target).addCallback(self.reportWhois, target, channel)


    def info_TIME(self, user, channel, args):
        """Usage: time - Gives the current system time."""
        self.reply(
            channel, 
            'The time at the tone will be: %s UTC.  ^G.' % (
                time.asctime(time.gmtime(time.time())),
            )
        )


    def seen_msg(self, user, channel):
        """Create the smartass response to a 'seen' query"""
        diff = time.time() - user.lastseen
        t = '(%s ago)' % formatTime(diff)
        if channel in user.channels:
            # If they spoke within the last 5 minutes, be a smartass
            if diff < 5 * 60:
                return 'Why, %s is right here!' % user.nick
            else:
                return 'Last heard %s: %s UTC %s' % (
                    user.nick, time.asctime(time.gmtime(user.lastseen)), t
                )
        else:
            return 'Last saw %s: %s UTC %s' % (
                user.nick, time.asctime(time.gmtime(user.lastseen)), t
            )


    def info_SEEN(self, user, channel, args):
        """Usage: seen <nick>[@network] - Reports the last time <nick> was seen"""
        target = args.strip().lower().split('@', 1)
        if len(target) == 1:
            bot = self
            target = target[0]
        else:
            target, network = target
            if not self.factory.allBots.has_key(network):
                self.reply(channel, '%s?  What is that?' % network)
                return
            bot = self.factory.allBots[network]
        if target == user.lower():
            self.reply(channel, "Why don't you tell me?")
        elif target == bot.nickname.lower():
            self.reply(channel, "Very funny.")
        else:
            user = bot.network.users.get(target, None)
            if user:
                self.reply(channel, bot.seen_msg(user, channel))
            else:
                self.reply(channel, "Never seen %s" % (args,))


    def info_OWNERS(self, user, channel, args):
        """Usage: owners - List the nicks who own me"""
        if len(args):
            try:
                net = self.factory.networkByAlias(args)
            except KeyError:
                self.reply(channel, "No such network")
            else:
                self.reply(channel, "My owners on %s: %s" % (net.alias, ', '.join(net.owners)))
        else:
            self.reply(channel, 'I am owned by: %s' % (', '.join(self.network.owners)))


    def info_SYNCS(self, user, channel, args):
        """Usage: syncs - Show synchronization configuration"""
        fmt = "%s is%s sync'd"
        items = self.network.syncs.items()
        self.reply(
            channel, 
            ', '.join([fmt % (x, not y and " not" or "") for (x, y) in items])
        )


    def info_BRIDGES(self, user, channel, args):
        """Usage: bridges - List the network bridges that are currently in place"""
        if not len(self.network.bridged):
            self.reply(channel, "I am not currently bridged to any network.")
        else:
            self.reply(
                channel,
                "From %s, I am bridged to: %s" % (
                    self.network.alias,
                    ', '.join(self.network.bridged)
                )
            )


    def info_NETWORKS(self, user, channel, args):
        """Usage: networks - List the networks I am currently connected to"""
        r = []
        for (net, bot) in self.factory.allBots.items():
            r.append('%s -> %s' % (net, bot.network.address))
        self.reply(channel, "Connected to: %s" % ', '.join(r))


    def info_WHO(self, user, channel, args):
        """Usage: who [channel@]<alias> - List users in the specified channel on the network with the specified alias"""
        args = args.split('@')
        if len(args) == 1:
            c, a = channel, args[0]
        elif len(args) >= 2:
            c, a = args[:2]
        else:
            self.reply(channel, 'Who!  Where?')
            return

        try:
            bot = self.factory.allBots[a]
        except KeyError:
            self.reply(channel, '%s?  What is that?' % (a,))
        else:
            bot.getUserList(c).addCallback(
                lambda users: ', '.join(users) or 'No one'
            ).addCallback(
                lambda users: self.reply(channel, '%d users on %s: %s' % (len(users.split(', ')), a, users))
            )


    # Priviledged commands
    def privinfo_GARBAGE(self, user, channel, args):
        self.reply(channel, "Garbage: %d objects" % (len(gc.garbage),))


    def privinfo_LOGLINES(self, user, channel, args):
        """Usage: loglines <integer> - Set the number of lines to check when starting or ending a conversation from a regexp."""
        try:
            lines = int(args)
        except:
            self.reply(channel, "%s is not a valid number of lines." % args)
        else:
            if lines < 0:
                self.reply(channel, "You must specify a positive number of lines.")
            else:
                self.LOG_LINES, old = lines, self.LOG_LINES
                self.reply(channel, "I will now examine %d lines of channel history (was examining %d lines)." % (lines, old))


    def privinfo_PREFIX(self, user, channel, args):
        """Usage: prefix [character] - Get or modify the prefix character"""
        args = args.strip()
        if len(args):
            self.PREFIX = args[0]
            self.reply(channel, 'Prefix changed to %s' % (self.PREFIX,))
        else:
            self.reply(channel, 'Prefix is %s' % (self.PREFIX,))


    def privinfo_ENABLE(self, user, channel, args):
        """Usage: enable <command> - Enable the specified disabled command"""
        args = args.strip().lower().split()
        for i in args:
            if i in self.network.disabled:
                self.network.disabled.remove(i)
                self.factory.setNetwork(self.network)
                self.reply(channel, i + ' enabled')
            else:
                self.reply(channel, i + ' is not disabled')


    def privinfo_DISABLE(self, user, channel, args):
        """Usage: disable <command> - Disable the specified enabled command"""
        args = args.strip().lower().split()
        for i in args:
            if i not in self.network.disabled:
                self.network.disabled.append(i)
                self.factory.setNetwork(self.network)
                self.reply(channel, i + ' disabled')
            else:
                self.reply(channel, i + ' is not enabled')

    def privinfo_SYNC(self, user, channel, args):
        """Usage: sync <topic|bans> <alias> - Synchronize channel state across the given bridge"""
        args = args.lower().strip()
        if args not in ('topic', 'bans'):
            self.reply(channel, "You can only sync 'topic' or 'bans'.")
        else:
            self.network.syncs[args] = not self.network.syncs.setdefault(args, 0)
            self.reply(channel, "%s %s." % (self.network.syncs[args] and "Enabled" or "Disabled", args))


    def privinfo_NICK(self, user, channel, args):
        """Usage: nick <new nick>[@<network>] - Change my nick"""
        args = args.split('@')
        if len(args) == 1:
            self.setNick(args[0])
        elif len(args) == 2:
            try:
                self.factory.allBots[args[1]].setNick(args[0])
            except KeyError:
                self.reply(channel, "Unknown network: %s" % (args[1],))
        else:
            raise error.UsageError


    def privinfo_REBUILD(self, user, channel, args):
        """Usage: rebuild [verbose] - Update my code from the modified source on disk"""
        doLog = args.strip().lower() == 'verbose'
        f = lambda m: rebuild.rebuild(m, doLog=doLog)
        soFar = []
        basicMods = [
            irc, basic, dirdbm, twistd, rebuild, reflect, copyright,
            info, util, secure, logview, shorter
        ]

        self.factory.finalizePlugins()
        try:
            for m in basicMods:
                soFar.append(m)
                f(m)
            
            for m in self.factory.pluginModules:
                f(m)
                for k in dir(m):
                    a = getattr(m, k)
                    if isinstance(a, types.ModuleType) and a not in soFar:
                        soFar.append(a)
                        f(a)

            self.factory.loadPlugins()
            
            self.rebuilds = self.rebuilds + 1
        except Exception, e:
            s = StringIO.StringIO()
            traceback.print_exc(file=s)
            self.reply(channel, 'An error occurred')
            self.reply(channel, s.getvalue())
        else:
            self.reply(channel, "Rebuilt.")


    def parseNetwork(self, args):
        p = args.split()
        if len(p) == 0:
            raise ValueError('Not enough arguments')
        elif len(p) == 1:
            return p
        elif len(p) > 2:
            raise ValueError('Too many arguments')
        else:
            alias = p[0]
            try:
                host, port = p[1].split(':', 1)
            except ValueError:
                host, port = p[1], '6667'
            try:
                port = int(port)
            except ValueError:
                raise ValueError("Invalid port number: %s" % port)
            return alias, host, port


    def privinfo_GBRIDGE(self, user, channel, args):
        """Usage: gbridge <alias> - Create a (two-way) bridge to the network with the specified alias from all other networks"""
        try:
            n = self.factory.networkByAlias(args)
            bot = self.factory.allBots[n.alias]
        except KeyError:
            self.reply(channel, "Not connected to %s." % args)
        else:
            for b in self.factory.allBots.values():
                if b is not bot:
                    b.createBridge(bot)
                    bot.createBridge(b)


    def createBridge(self, other):
        if other.network.alias not in self.network.bridged:
            self.network.bridged.append(other.network.alias)
            self.factory.setNetwork(self.network)
            for ch in self.network.channels:
                self.msg(ch, "Created bridge to %s (%s:%d)." % ((other.network.alias,) + other.network.address))


    def privinfo_BRIDGE(self, user, channel, args):
        """Usage: bridge <alias> - Create a (two-way) bridge to the network with the specified alias"""
        try:
            n = self.factory.networkByAlias(args)
        except KeyError:
            self.reply(channel, "No such network.")
        else:
            if n.alias == self.network.alias:
                self.reply(channel, "No need to bridge to ourself, captain.")
            elif n.alias in self.network.bridged:
                self.reply(channel, "I'm already bridged to %s." % n.alias)
            else:
                self.createBridge(self.factory.allBots[n.alias])
                self.factory.allBots[n.alias].createBridge(self)


    def privinfo_UNBRIDGE(self, user, channel, args):
        """Usage: unbridge <alias> - Destroy a (two-way) bridge to the network with the specified alias"""
        if args in self.network.bridged:
            self.network.bridged.remove(args)
            self.factory.setNetwork(self.network)
            self.msg(channel, "Unbridged from %s." % (args,))
            try:
                bot = self.factory.allBots[args]
                bot.network.bridged.remove(self.network.alias)
                self.factory.setNetwork(bot.network)
                bot.msg(channel, "Unbridged from %s" % (args,))
            except (ValueError, KeyError), e:
                pass
        else:
            self.reply(channel, "I'm not bridged to %s." % (args,))
    
    
    def info_CONVERSATION(self, user, channel, args):
        """Usage: conversation <begin | end> <name> [! <regexp>] - Create a new conversation tag or close a previously opened conversation."""
#        if channel == user:
#            self.reply(channel, "You cannot do that in a private query.")
#            return
        try:
            cmd, rest = args.split(None, 1)
        except ValueError:
            self.reply(channel, "Incorrect syntax - see 'help conversation'")
            return

        if cmd.lower() not in ('begin', 'end'):
            self.reply(channel, "Incorrect syntax - see 'help conversation'")
            return

        try:
            name, regex = rest.split(' ! ')
        except ValueError:
            name = rest
            when = time.time()
            whenLine = None
        else:
            try:
                regex = re.compile(regex)
            except re.error:
                self.reply(channel, 'Illegal regular expression')
                return

            lines = self.factory.getLogsFor(channel)[-self.LOG_LINES:]
            lines.reverse()
            for l in lines:
                s = '<' + l[2] + '> ' + l[3]
                if regex.search(s):
                    when = l[0]
                    whenLine = s
                    break
            else:
                self.reply(channel, 'Expression not found (examined %d lines).' % self.LOG_LINES)
                return
        
        if cmd.lower() == 'begin':
            conv = Conversation(user, channel, name, when)
            self.factory.addConversation(conv)
            self.reply(
                channel,
                "Beginning tagged conversation %r%s" % (
                    name, whenLine and (" from %r." % whenLine) or "."
                )
            )
        else:
            try:
                conv = self.factory.findConversation(channel, name.lower())
            except ValueError:
                self.reply(channel, "No conversation found matching %r" % rest)
            else:
                if user != conv.recordedBy:
                    if self.network.findOwner(user) == -1:
                        self.reply(channel, "You cannot do that.")
                    else:
                        def endConvo(result):
                            if result:
                                self.factory.endConversation(conv, when)
                                self.reply(
                                    channel, "Ended conversation tagged %r%s" % (
                                        conv.name, whenLine and (" at %r." % whenLine) or "."
                                    )
                                )
                            else:
                                self.reply(channel, "You are not authorized to do that.")
                        self.getComponent(IAuthenticator).authenticate(user).addCallback(endConvo)
                elif when < conv.started:
                    self.reply(channel, "You cannot end a conversation before it has begun.")
                elif conv.ended:
                    self.reply(channel, "That conversation has already been ended.")
                else:
                    self.factory.endConversation(conv, when)
                    self.reply(
                        channel, "Ended tagged conversation %r%s" % (
                            name, whenLine and (" at %r." % whenLine) or "."
                        )
                    )


    def info_CONVERSATIONS(self, user, channel, args):
        """Usage: conversations - Show all currently open conversations"""
        l = [c for c in self.factory.data['conversations'] if not c.ended]
        if not len(l):
            self.reply(channel, "There are currently no open conversations.")
            return

        self.reply(
            channel, 
            '\n'.join([
                'Channel: %s  Name: %s' % (c.channel, c.name) for c in l
            ])
        )


    def privinfo_RESPOND(self, user, channel, args):
        """Usage: respond <response> - Respond to a secure command challenge"""
        response = self.factory.decrypt(args)
        for i in range(len(self.secureList)):
            if self.secureList[i].verifyResponse(response):
                try:
                    self.secureList[i].invoke()
                except error.UsageError, e:
                    self.reply(user, str(e))
                except Exception, e:
                    s = StringIO.StringIO()
                    traceback.print_exc(file=s)
                    self.reply(user, "An unexpected error occurred:\n" + s.getvalue())
                del self.secureList[i]
                break
        else:
            self.reply(channel, "Response unrecognized.")


    def secure_OWNER(self, user, channel, args):
        """Usage: owner <nick>[@<network>] - Add or remove the given nick from my owner list"""
        try:
            target, net = args.split('@', 1)
        except ValueError:
            target, net = args.strip(), self.network.alias
        try:
            net = self.factory.networkByAlias(net)
        except KeyError:
            self.reply(channel, "No such network.")
            return

        if not target:
            self.reply(channel, "Toggle whom on the owner list?")
        elif target not in net.owners:
            net.owners.append(target)
            self.factory.setNetwork(net)
            self.reply(channel, "Added %s to owner list." % (target,))
        else:
            i1, i2 = net.findOwner(target), net.findOwner(user)
            if i1 < i2:
                self.reply(channel, "That owner is more senior than you.")
            elif i1 == i2:
                self.reply(channel, "You cannot remove yourself from the owner list.")
            else:
                net.owners.remove(target)
                self.factory.setNetwork(net)
                self.reply(channel, "Removed %s from owner list." % (target,))


    def secure_DEBUG(self, user, channel, args):
        """Usage: debug - Toggle debug flag"""
        self.debug = not self.debug
        self.reply(channel, "%s debugging." % (self.debug and "Enabled" or "Disabled"))


    def secure_UPGRADE(self, user, channel, args):
        pass


    # CTCP
    def ctcpQuery_VERSION(self, user, channel, data):
        if data is not None:
            self.quirkyMessage("Why did %s send '%s' with a VERSION query?"
                               % (user, data))

        nick = string.split(user,"!")[0]
        self.ctcpMakeReply(nick, [('VERSION', '%s' % self.versionName)])


class Network:
    def __init__(self, root, alias, address, nickname, owners = None, channels = None):
        self.alias, self.address, self.nickname = alias, address, nickname
        self.owners = owners or util.CaseList()
        self.passwords = util.CaseDict()
        self.channels = channels or util.CaseList()
        self.disabled = util.CaseList()

        p = os.path.join(root, alias)
        if not os.path.isdir(p):
            os.mkdir(p)

        self.users = dirdbm.Shelf(os.path.join(root, alias, 'users'))
        self.bridged = util.CaseList()
        self.syncs = util.CaseDict()


    def findOwner(self, name):
        try:
            return self.owners.index(name)
        except ValueError:
            return -1


    def __repr__(self):
        return 'Network(%r, %r)' % (self.alias, self.address)


class ThrottledClientFactory(protocol.ClientFactory):
    lostDelay = 2
    failedDelay = 60


    def clientConnectionLost(self, connector, reason):
        reactor.callLater(self.lostDelay, connector.connect)


    def clientConnectionFailed(self, connector, reason):
        reactor.callLater(self.failedDelay, connector.connect)


class InfoBotFactory(ThrottledClientFactory, styles.Versioned, components.Componentized):
    """A factory for InfoBots."""
    
    protocol = InfoBot
    allBots = None
    logCache = None
    plugins = None
    pluginModules = None

    webhost = None
    webport = None

    inBytes = 0
    outBytes = 0
    
    startTime = None
    createTime = None

    persistenceVersion = 1
    
    authenticator = ServicesAuthenticator

    def __init__(self, dbpath, webhost = None, webport = None, bits = 128):
        components.Componentized.__init__(self)
        self.createTime = time.time()

        self.dbpath = dbpath

        self.shortenMap = {}
        self.logCache = util.CaseDict()
        self.allBots = {}

        self.logs = util.CaseDict()
        self.webhost, self.webport = webhost, webport

        self.data = dirdbm.Shelf(dbpath)
        self.data['conversations'] = []
        self.data['networks'] = dirdbm.Shelf(os.path.join(dbpath, 'networks'))
        if RSA:
            self.data['key'] = RSA.generate(bits, randsource)
        else:
            self.data['key'] = None

        self.data['factoids'] = dirdbm.Shelf(os.path.join(self.dbpath, 'factoids'))
        for a in range(ord('a'), ord('z') + 1):
            self.data['factoids'][chr(a)] = dirdbm.Shelf(os.path.join(self.dbpath, 'factoids', chr(a)))
        self.data['factoids']['other'] = dirdbm.Shelf(os.path.join(self.dbpath, 'factoids', 'other'))


    def createNetwork(self, address, alias, nickname, owners = None, channels = None):
        self.setNetwork(Network(self.dbpath, alias, address, nickname, owners, channels))


    def setNetwork(self, net):
        nets = self.data['networks']
        nets[net.alias] = net
        self.data['networks'] = nets
        for (alias, bot) in self.allBots.items():
            if alias == net.alias:
                bot.network = net


    def setAuthenticator(self, auth):
        self.authenticator = auth


    def networkByAlias(self, alias):
        for n in self.data['networks'].values():
            if n.alias == alias:
                return n
        raise KeyError, alias


    def addConversation(self, conv):
        convs = self.data['conversations']
        convs.append(conv)
        self.data['conversations'] = convs


    def findConversation(self, channel, name):
        convs = self.data['conversations']
        convs.reverse()
        for c in convs:
            if c.channel == channel and c.name.lower() == name.lower():
                return c
        raise ValueError, "No such conversation"


    def endConversation(self, convo, when = None):
        l = self.data['conversations']
        i = l.index(convo)
        if when is None:
            l[i].ended = time.time()
        else:
            l[i].ended = when
        self.data['conversations'] = l


    def setFact(self, fact, value):
        if fact:
            fact = fact.lower()
            if fact[0].isalpha():
                dict = self.data['factoids'][fact[0]]
            else:
                dict = self.data['factoids']['other']
            dict[fact] = value


    def unsetFact(self, fact):
        fact = fact.lower()
        if fact:
            if fact[0].isalpha():
                dict = self.data['factoids'][fact[0]]
            else:
                dict = self.data['factoids']['other']
            del dict[fact]


    def lookupFact(self, fact):
        if fact:
            fact = fact.lower()
            if fact[0].isalpha():
                dict = self.data['factoids'][fact[0]]
            else:
                dict = self.data['factoids']['other']
            return dict.get(fact, None)


    def countFactoids(self):
        f = len(self.data['factoids']['other'])
        for k in map(chr, range(ord('a'), ord('z') + 1)):
            f = f + len(self.data['factoids'][k])
        return f


    def log(self, network, channel, user, message):
        when = time.time()
        channel = channel.lower()
        if not self.logs.has_key(channel):
            self.logs[channel] = file(os.path.join('logs', util.safeFilename(channel)), 'a')
        for line in message.splitlines():
            msg = '%f %s %s %s' % (when, network, user, line)
            self.logs[channel].write(msg + '\n')
            when += 0.01
        self.logs[channel].flush()


    def getLogsFor(self, channel):
        if self.logs.has_key(channel):
            self.logs[channel].flush()
        logfile = os.path.join('logs', util.safeFilename(channel))
        if os.path.exists(logfile):
            input = file(logfile)
            lines = input.readlines()
            input.close()
            lines = [l.split(None, 3) for l in lines]
            lines = [[float(l[0])] + l[1:] for l in lines]
            return lines
        return []


    def encrypt(self, challenge):
        if self.data['key']:
            return secure.mangle(self.data['key'].encrypt(challenge, '')[0])
        raise RuntimeError, "PyCrypto is not installed"

    def decrypt(self, response):
        if self.data['key']:
            return self.data['key'].decrypt(secure.demangle(response))
        raise RuntimeError, "PyCrypto is not installed"


    def loadPlugins(self):
        # It would be wonderful if we could use this code
        # but it breaks rebuild() and possibly other things

        #path = python.util.sibpath(__file__, 'plugins/plugins.tml')
        #deferredGetPlugIns(
        #    'infobot',
        #    [path]
        #).addCallbacks(self._gotPlugins, self._errPlugins)
        
        path = python.util.sibpath(__file__, 'plugins')
        if path not in sys.path:
            sys.path.append(path)
        self._gotPlugins(plugin.getPlugIns('infobot'))

    def _gotPlugins(self, pluginList):
        del self.pluginModules[:]
        for p in pluginList:
            try:
                module = p.load()
            except:
                log.msg("Failed to load plugin: " + p.module)
                log.deferr()
            else:
                if hasattr(module, 'init'):
                    try:
                        module.init(self)
                    except:
                        log.msg("Failed to initialize plugin: " + module.__name__)
                        log.deferr()
                    else:
                        log.msg("Loaded and initialized " + p.module)
                        self.pluginModules.append(module)
                else:
                    log.msg("Loaded " + p.module)
                    self.pluginModules.append(module)

        self.plugins.clear()
        for prefix in ('info_', 'privinfo_', 'secure_', 'filter_'):
            l = self.plugins[prefix] = {}
            for m in self.pluginModules:
                l.update(dict([(x[len(prefix):], getattr(m, x)) for x in dir(m) if x.startswith(prefix)]))


    def _errPlugins(self, failure):
        failure.printTraceback()
        del self.pluginModules[:]
        self.plugins.clear()


    def buildProtocol(self, address):
        address = address[1:]   # Nuke the protocol

        for n in self.data['networks'].values():
            if n.address == address:
                break
        else:
            log.msg("Unknown network address: " + repr(address))
            return InstantDisconnectProtocol()

        p = self.protocol(n, self.plugins)
        p.setAdapter(IAuthenticator, self.authenticator)
        self.allBots[n.alias] = p
        p.factory = self
        return p
    
    
    def clientConnectionLost(self, connector, reason):
        for n in self.data['networks'].values():
            if connector.getDestination()[1:] == n.address:
                if self.allBots.has_key(n.alias):
                    if not self.allBots[n.alias].hasQuit:
                        ThrottledClientFactory.clientConnectionLost(self, connector, reason)
                    self.inBytes += self.allBots[n.alias].inBytes
                    self.outBytes += self.allBots[n.alias].outBytes
                    del self.allBots[n.alias]
                    return
        log.msg("Not reconnecting client to %r" % (connector.getDestination(),))


    def startFactory(self):
        L = []
        for conv in self.data['conversations']:
            L.append(conv)
            if isinstance(conv.started, float):
                continue
            conv.started = time.mktime(conv.started) + time.timezone
            if conv.ended is not None:
                conv.ended = time.mktime(conv.ended) + time.timezone
        self.data['conversations'] = L


        self.plugins = {}
        self.pluginModules = []
        self.loadPlugins()

        self.allBots = {}
        self.starttime = time.time()

        if not os.path.exists('logs'):
            os.mkdir('logs')
        ThrottledClientFactory.startFactory(self)


    def stopFactory(self):
        del self.allBots, self.plugins
        for f in self.logs.values():
            f.close()
        self.logs = {}

        self.finalizePlugins()
        del self.pluginModules[:]
        ThrottledClientFactory.stopFactory(self)

    def finalizePlugins(self):
        for p in self.pluginModules:
            if hasattr(p, 'fini'):
                try:
                    p.fini(self)
                except:
                    log.msg("Exception in finalizer for " + p.__name__)
                    log.deferr()
                else:
                    log.msg('Finalized ' + p.__name__)

    def upgradeToVersion1(self):
        import binascii
        self.createTime = time.mktime(self.createTime) + time.timezone
        for path in os.listdir('logs'):
            newf = util.safeFilename(binascii.unhexlify(path))
            newp = os.path.join('logs', newf)
            os.rename(os.path.join('logs', path), newp)
            f = file(newp)
            o = file(newp + '.new', 'w')
            for line in f:
                parts = line.split(None, 1)
                when = float(parts[0]) + time.timezone
                o.write('%0.2f %s' % (when, parts[1]))
            f.close()
            o.close()
            os.rename(newp + '.new', newp)
        L = []
        for conv in self.data['conversations']:
            conv.started = time.mktime(conv.started) + time.timezone
            if conv.ended is not None:
                conv.ended = time.mktime(conv.ended) + time.timezone
            L.append(conv)
        self.data['conversations'] = L


    # Fuck me.
    def __getstate__(self):
        d = styles.Versioned.__getstate__(self)
        d['authenticator'] = reflect.qual(self.authenticator)
        return d

    # Fuck me with a rake.
    def __setstate__(self, state):
        auth = state.get('authenticator', self.authenticator)
        if isinstance(auth, str):
            state['authenticator'] = reflect.namedObject(auth)
        styles.Versioned.__setstate__(self, state)


components.registerAdapter(
    adapters.SchedulerAdapter, InfoBotFactory, interfaces.IScheduler
)

components.registerAdapter(
    adapters.StorageAdapter, InfoBotFactory, interfaces.IStorage
)
