
# Copyright (C) 2002 Jp Calderone.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# System imports
import time, cgi, urllib

# Twisted imports
from twisted.web import resource, static
from twisted.web import util as webutil

# Sibling imports
import util


class LogWidget(resource.Resource):
    """A Resource that displays a conversation."""

    FMT = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">
<html><head><link href="/logview.css" type="text/css" rel="stylesheet"></head>
<body>%s</body></html>
"""
    LINE_FMT = """
<div class="ircLine">
  <span class="timestamp">%s</span>&lt;
  <span class="identity">
    <span class="nick">%s</span>
    <span class="network">@%s</span>&gt;
  </span>
  <span class="%s">%s</span>
</div>"""

    def __init__(self, factory, convo):
        self.factory = factory
        self.conversation = convo


    def lookupLines(self):
        return util.findConversation(
            self.conversation,
            self.factory.getLogsFor(self.conversation.channel)
        )


    def format(self, when, network, user, message):
        s = self.lineNo % 2 and 'message' or 'altmessage'
        self.lineNo += 1
        return self.LINE_FMT % (
            time.ctime(when), user, network, s, cgi.escape(message)
        )


    def render(self, request):
        """
        Return the text of a tagged conversation.
        """
        lines = self.lookupLines()

        r = []
        self.lineNo = 0
        for i in range(0, len(lines), 2):
            r.append(self.format(*lines[i]))
            if i + 1 < len(lines):
                r.append(self.format(*lines[i + 1]))
        return self.FMT % '\n'.join(r)


class LogViewer(resource.Resource):
    """I am a Resource that displays part a list of IRC logs."""


    def __init__(self, factory, channel):
        self.factory = factory
        self.channel = channel
        self.children = {}
    
    
    def getChild(self, path, request):
        path = urllib.unquote(path)
        if path == '':
            return webutil.Redirect(request.path[:-1])
        try:
            conv = self.factory.findConversation(self.channel, path)
        except ValueError:
            return resource.Resource.getChild(self, urllib.quote(path), request)
        else:
            return LogWidget(self.factory, conv)


    def render(self, request):
        convos = [c.name for c in self.factory.data['conversations'] if c.channel == self.channel]
        l = []
        for c in convos:
            l.append(
                '<a href="%s/%s">%s</a>' % (
                    urllib.quote(self.channel),
                    urllib.quote(c), c
                )
            )
        return '<br>'.join(l)


class ChannelViewer(resource.Resource):
    """I am a Resource that displays a list channels."""


    def __init__(self, factory):
        self.factory = factory
        self.children = {}


    def getChild(self, path, request):
        channels = dict([(c.channel, 0) for c in self.factory.data['conversations']])
        path = urllib.unquote(path)
        if path == '':
            return self
        elif path in channels:
            return LogViewer(self.factory, path)
        elif path == 'logview.css':
            return static.File('logview.css')
        else:
            return resource.Resource.getChild(self, urllib.quote(path), request)


    def render(self, request):
        channels = dict([(c.channel, 0) for c in self.factory.data['conversations']])
        l = []
        for c in channels.keys():
            l.append('<a href="%s">%s</a>' % (urllib.quote(c), c))
        return '<br>'.join(l)
